class Text2ImageDefaultParams:
    style = "DEFAULT"
    width = 1024
    height = 1024
    art_gpt = False
    prompt = "Cat"
    negative_prompt = ""
    num_images = 1

    async def comb(
            self,
            style: str,
            width: int,
            height: int,
            art_gpt: bool,
            prompt: str,
            negative_prompt: str,
            num_images: str
    ) -> list:
        _style = style or self.style
        _width = width or self.width
        _height = height or self.height
        _art_gpt = art_gpt or self.art_gpt
        _prompt = prompt or self.prompt
        _negative_prompt = negative_prompt or self.negative_prompt
        _num_images = num_images or self.num_images

        return [
            {
                "type": "GENERATE",
                "style": _style,
                "width": _width,
                "height": _height,
                "numImages": _num_images,
                "censor": {"useGigaBeautificator": art_gpt},
                "generateParams": {"query": _prompt},
                "negativePromptDecoder": _negative_prompt
            }
        ]


class Text2AnimationDefaultParams:
    prompts = ["Cat", "Cat"]
    negative_prompts = ["", ""]
    directions = ["LEFT_SINUS_Y", "LEFT_SINUS_Y"]
    width = 640
    height = 640

    async def comb(
            self,
            prompts: list[str],
            negative_prompts: list[str],
            directions: list[str],
            width: int,
            height: int
    ) -> list:
        _width = self.width if width is None else width
        _height = self.height if height is None else height
        _directions = self.directions if directions is None else directions
        _prompts = self.prompts if prompts is None else prompts
        _negative_prompts = self.negative_prompts if negative_prompts is None else negative_prompts

        return [
            {
                "width": _width,
                "height": _height,
                "animation_steps": [
                    {
                        "prompt": _prompts[x],
                        "negative_prompt": _negative_prompts[x],
                        "direction": _directions[x],
                        "acceleration": "1"
                    }
                    for x in range(2)
                ]
            }
        ]


class Text2VideoDefaultParams:
    prompt = "Cat"
    width = 512
    height = 512

    async def comb(
            self,
            prompt: str,
            width: int,
            height: int
    ) -> list:
        _width = self.width if width is None else width
        _height = self.height if height is None else height
        _prompt = self.prompt if prompt is None else prompt

        return [
            {
                "width": _width,
                "height": _height,
                "prompt": _prompt
            }
        ]



from __future__ import annotations

import asyncio
import base64
import json
import time
from io import BytesIO

import aiohttp

from .API_types import ApiApi, ApiWeb


class FusionBrainApi:
    def __init__(self, api: ApiApi | ApiWeb):
        if hasattr(api, "type"):
            if api.type not in [ApiApi.type, ApiWeb.type]:
                raise TypeError("Invalid API type")
            else:
                self.api = api
        else:
            raise TypeError("Invalid API structure")

    async def get_styles(self) -> dict:
        async with aiohttp.ClientSession() as session:
            n_url = self.api.urls.url_get_styles
            async with session.get(n_url) as response:
                return await response.json()

    async def text2image(
            self,
            prompt: str,
            negative_prompt: str | None = None,
            style: str | None = None,
            width: int | None = None,
            height: int | None = None,
            art_gpt: bool | None = None,
            num_images: int | None = None,

            pipeline_id: int = 0, # currently: 0 == "Kandinsky 3.1", 1 == "Kandinsky 3.0", 2 == "Kandinsky 3.2"
            max_time: int = 2 * 60,  # max time generation on seconds (after return error)
            sleep_sec: int = 4
    ) -> BytesIO:
        params = await self.api.text2image_default_params.comb(
            style, width, height, art_gpt, prompt, negative_prompt, num_images
        )

        data = aiohttp.FormData()
        data.add_field("params",
                       json.dumps(params[0]),
                       content_type="application/json",
                       )

        async with aiohttp.ClientSession(headers=await self.api.get_headers()) as session:
            pipelines_url = self.api.urls.url_pipelines
            async with session.get(f"{pipelines_url}?type=TEXT2IMAGE") as resp:
                pipelines = await resp.json()

            n_url = self.api.urls.url_texts2image_run
            if self.api.type == ApiApi.type:
                data.add_field(
                    "pipeline_id",
                    pipelines[pipeline_id]['id'],
                )
                async with session.post(f"{n_url}", data=data) as resp:
                    result = await resp.json()
            else:
                async with session.post(f"{n_url}?pipeline_id={pipelines[pipeline_id]['id']}", data=data) as resp:
                    result = await resp.json()
        if "error" in result:
            raise ValueError(result)

        uuid = result["uuid"]
        first_sleep_sec = result["status_time"] if "status_time" in result else 4
        return await self.polling(uuid, max_time, "img", sleep_sec, first_sleep_sec)

    async def text2animation(
            self,
            prompts: list[str],
            negative_prompts: list[str] | None = None,
            directions: list[str] | None = None,
            width: int | None = None,
            height: int | None = None,

            max_time: int = 5 * 60,  # max time generation on seconds (after return error)
            sleep_sec: int = 4
    ) -> BytesIO:
        if self.api.type != "web":
            raise TypeError("text2animation only supports web")

        params = await self.api.text2animation_default_params.comb(
            prompts, negative_prompts, directions, width, height
        )
        params = params[0]

        data = aiohttp.FormData()
        data.add_field("params",
                       json.dumps(params),
                       content_type="application/json",
                       )

        async with aiohttp.ClientSession(headers=await self.api.get_headers()) as session:
            pipelines_url = self.api.urls.url_pipelines
            async with session.get(f"{pipelines_url}?type=ANIMATION") as resp:
                pipelines = await resp.json()

            n_url = self.api.urls.url_text2animation_run
            async with session.post(f"{n_url}?pipeline_id={pipelines[0]['id']}", data=data) as resp:
                result = await resp.json()

        if "error" in result:
            raise ValueError(result)

        uuid = result["uuid"]
        first_sleep_sec = result["status_time"] if "status_time" in result else 4
        return await self.polling(uuid, max_time, "anim", sleep_sec, first_sleep_sec)

    async def text2video(
            self,
            prompt: str,
            width: int | None = None,
            height: int | None = None,

            max_time: int = 6 * 60,  # max time generation on seconds (after return error)
            sleep_sec: int = 4
    ) -> BytesIO:
        if self.api.type != ApiWeb.type:
            raise TypeError("text2video only supports web")

        params = await self.api.text2video_default_params.comb(
            prompt, width, height
        )
        params = params[0]

        data = aiohttp.FormData()
        data.add_field("params",
                       json.dumps(params),
                       content_type="application/json",
                       )

        async with aiohttp.ClientSession(headers=await self.api.get_headers()) as session:
            pipelines_url = self.api.urls.url_pipelines
            async with session.get(f"{pipelines_url}?type=TEXT2VIDEO") as resp:
                pipelines = await resp.json()

            n_url = self.api.urls.url_text2video_run
            async with session.post(f"{n_url}?pipeline_id={pipelines[0]['id']}", data=data) as resp:
                result = await resp.json()

        if "error" in result:
            raise ValueError(result)

        uuid = result["uuid"]
        first_sleep_sec = result["status_time"] if "status_time" in result else 4
        return await self.polling(uuid, max_time, "video", sleep_sec, first_sleep_sec)

    async def polling(self, uuid: str, max_time: int, type_generation: str, sleep_sec = 4, first_sleep_sec = 4) -> BytesIO:
        start_time = time.time()
        while time.time() - (start_time + max_time) < 0:
            async with aiohttp.ClientSession(headers=await self.api.get_headers()) as session:
                if type_generation == "img":
                    n_url = self.api.urls.url_text2image_status
                elif type_generation == "anim":
                    n_url = self.api.urls.url_text2animation_status
                elif type_generation == "video":
                    n_url = self.api.urls.url_text2video_status
                else:
                    raise TypeError("type_generation must be 'img' or 'anim' or 'video'")
                n_url = n_url.replace("$uuid", uuid)
                async with session.get(n_url) as resp:
                    result = await resp.json()
                if result["status"] == "DONE":
                    censored = result["result"]["censored"] if self.api.type == ApiApi.type else result["censored"]
                    if isinstance(censored, bool) and censored or isinstance(censored, list) and any(censored):
                        raise ValueError("censored: is True")
                    else:
                        if type_generation == "img":
                            if self.api.type == ApiApi.type:
                                return BytesIO(base64.b64decode(result["result"]["files"][0]))
                            else:
                                async with session.get(result["images"][0]) as resp_img:
                                    if resp_img.status == 200:
                                        return BytesIO(await resp_img.read())
                                    else:
                                        raise ValueError("Fail install image from url")
                        elif type_generation in ["anim", "video"]:
                            return BytesIO(base64.b64decode(result["video"]))
                        else:
                            raise TypeError("type_generation must be 'img' or 'anim' or 'video'")
                elif result["status"] == "FAIL":
                    raise ValueError(f"status is FAIL: {result['status']}")

            await asyncio.sleep(sleep_sec)

        raise ValueError(f"timeout: {max_time} seconds")



class BaseUrls:
    url_get_styles_base = "https://cdn.fusionbrain.ai/static/styles"
    url_api_base = "https://api-key.fusionbrain.ai/key/api/v1"
    url_web_base = "https://api.fusionbrain.ai/web/api/v1"
    url_login_base = "https://fusionbrain.ai/api/auth"

class ApiUrls:
    url_get_styles = f"{BaseUrls.url_get_styles_base}/key"

    url_pipelines = f"{BaseUrls.url_api_base}/pipelines"

    url_text2image_run = f"{BaseUrls.url_api_base}/pipeline/run"
    url_text2image_status = f"{BaseUrls.url_api_base}/pipeline/status/$uuid"

class WebUrls:
    url_csrf = f"{BaseUrls.url_login_base}/csrf/"
    url_keycloak = f"{BaseUrls.url_login_base}/signin/keycloak/"
    url_session = f"{BaseUrls.url_login_base}/session/"
    url_check_token = f"{BaseUrls.url_web_base}/text2image/availability"

    url_get_styles = f"{BaseUrls.url_get_styles_base}/web"

    url_pipelines = f"{BaseUrls.url_web_base}/pipelines"

    url_text2image_run = f"{BaseUrls.url_web_base}/text2image/run"
    url_text2image_status = f"{BaseUrls.url_web_base}/text2image/status/$uuid"

    url_text2animation_run = f"{BaseUrls.url_web_base}/animation/run"
    url_text2animation_status = f"{BaseUrls.url_web_base}/animation/status/$uuid"

    url_text2video_run = f"{BaseUrls.url_web_base}/text2video/run"
    url_text2video_status = f"{BaseUrls.url_web_base}/text2video/status/$uuid"